/*  Emacs style mode select   -*- C++ -*-  */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id:$ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  $Log:$ */
/*  */
/*  DESCRIPTION: */
/* 	DOOM graphics stuff for X11, UNIX. */
/*  */
/* ----------------------------------------------------------------------------- */

static const char
rcsid[] = "$Id: i_x.c,v 1.6 1997/02/03 22:45:10 b1 Exp $";

#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <X11/extensions/XShm.h>
/*  Had to dig up XShm.c for this one. */
/*  It is in the libXext, but not in the XFree86 headers. */
#ifdef LINUX
int XShmGetEventBase( Display* dpy ); /*  problems with g++? */
#endif

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <errnos.h>
#include <signal.h>

#include "i_system.h"
/* #include "v_video.h" */

/* #include "doomdef.h" */
#include "r_local.h"
#include "am_map.h"

#define POINTER_WARP_COUNTDOWN	1

Display*	X_display=0;
Window		X_mainWindow;
Colormap	X_cmap;
Visual*		X_visual;
GC		X_gc;
XEvent		X_event;
Screen		*X_screen;
XVisualInfo	X_visualinfo;

XImage*		image;
int		X_width;
int		X_height;
int		bpp;

int	fbnum=0;

int		pixel_size;
lighttable_t	truecolor_palette[256];
extern byte	*orig_colormaps;

extern	unsigned long fuzzmask;	/*  for fuzz effect in 15/16 bit mode */

/*  MIT SHared Memory extension. */
boolean		doShm;

XShmSegmentInfo	X_shminfo;
int		X_shmeventtype;

/*  Fake mouse handling. */
/*  This cannot work properly w/o DGA. */
/*  Needs an invisible mouse cursor at least. */
boolean		grabMouse;
int		doPointerWarp = POINTER_WARP_COUNTDOWN;

/*  */
/*   Translates the key currently in X_event */
/*  */

int xlatekey(void)
{
	int	rc;

	rc = XKeycodeToKeysym(X_display, X_event.xkey.keycode, 0);

	switch(rc)
	{
		case XK_BackSpace:
			rc = KEY_BACKSPACE;	break;
		case XK_Tab:
			rc = KEY_TAB;		break;
		case XK_Return:
		case XK_KP_Enter:
			rc = KEY_ENTER;		break;
		case XK_Pause:
			rc = KEY_PAUSE;		break;
		case XK_Escape:
			rc = KEY_ESCAPE;	break;

      case XK_Right:	rc = KEY_RIGHTARROW;	break;
      case XK_Left:	rc = KEY_LEFTARROW;	break;
      case XK_Up:	rc = KEY_UPARROW;	break;
      case XK_Down:	rc = KEY_DOWNARROW;	break;

	case XK_Insert:		rc=KEY_INSERT;	break;
	case XK_Home:		rc=KEY_HOME;	break;
	case XK_Page_Up:	rc=KEY_PAGEUP;	break;
	case XK_Delete:		rc=KEY_DELETE;	break;
	case XK_End:		rc=KEY_END;	break;
	case XK_Page_Down:	rc=KEY_PAGEDOWN;break;

      case XK_F1:	rc = KEY_F1;		break;
      case XK_F2:	rc = KEY_F2;		break;
      case XK_F3:	rc = KEY_F3;		break;
      case XK_F4:	rc = KEY_F4;		break;
      case XK_F5:	rc = KEY_F5;		break;
      case XK_F6:	rc = KEY_F6;		break;
      case XK_F7:	rc = KEY_F7;		break;
      case XK_F8:	rc = KEY_F8;		break;
      case XK_F9:	rc = KEY_F9;		break;
      case XK_F10:	rc = KEY_F10;		break;
      case XK_F11:	rc = KEY_F11;		break;
      case XK_F12:	rc = KEY_F12;		break;
	

      case XK_KP_Equal:
      case XK_equal:	rc = KEY_EQUALS;	break;

      case XK_KP_Subtract:
      case XK_minus:	rc = KEY_MINUS;		break;

      case XK_Shift_L:
      case XK_Shift_R:
	rc = KEY_RSHIFT;
	break;
	
      case XK_Control_L:
      case XK_Control_R:
	rc = KEY_RCTRL;
	break;
	
      case XK_Alt_L:
      case XK_Meta_L:
      case XK_Alt_R:
      case XK_Meta_R:
	rc = KEY_RALT;
	break;

/* 	case XK_KP_Insert:	rc='0';	break; */
	case XK_KP_End:		rc='1';	break;
	case XK_KP_Down:	rc='2';	break;
	case XK_KP_Page_Down:	rc='3';	break;
	case XK_KP_Left:	rc='4';	break;
	case XK_KP_Begin:	rc='5';	break;
	case XK_KP_Right:	rc='6';	break;
	case XK_KP_Home:	rc='7';	break;
	case XK_KP_Up:		rc='8';	break;
/* 	case XK_KP_Page_Up:	rc='9';	break; */

	case XK_KP_Divide:	rc='/';	break;
	case XK_KP_Multiply:	rc='*';	break;
	
      default:

	if (rc >= XK_space && rc <= XK_asciitilde)
	    rc = rc - XK_space + ' ';
	if (rc >= 'A' && rc <= 'Z')
	    rc = rc - 'A' + 'a';
	break;
    }
    return rc;
}

void I_ShutdownGraphics(void)
{
  /*  Detach from X server */
  if (!XShmDetach(X_display, &X_shminfo))
	    I_Error("XShmDetach() failed in I_ShutdownGraphics()");

  /*  Release shared memory. */
  shmdt(X_shminfo.shmaddr);
  shmctl(X_shminfo.shmid, IPC_RMID, 0);

  /*  Paranoia. */
  image->data = NULL;
}

static int	lastmousex = 0;
static int	lastmousey = 0;
boolean		mousemoved = false;
boolean		shmFinished;

void I_GetEvent(void)
{

    event_t event;

    /*  put event-grabbing stuff in here */
    XNextEvent(X_display, &X_event);
    switch (X_event.type)
    {
      case KeyPress:
	event.type = ev_keydown;
	event.data1 = xlatekey();
	H2_PostEvent(&event);
	/*  fprintf(stderr, "k"); */
	break;
      case KeyRelease:
	event.type = ev_keyup;
	event.data1 = xlatekey();
	H2_PostEvent(&event);
	/*  fprintf(stderr, "ku"); */
	break;
      case ButtonPress:
	event.type = ev_mouse;
	event.data1 =
	    (X_event.xbutton.state & Button1Mask)
	    | (X_event.xbutton.state & Button2Mask ? 2 : 0)
	    | (X_event.xbutton.state & Button3Mask ? 4 : 0)
	    | (X_event.xbutton.button == Button1)
	    | (X_event.xbutton.button == Button2 ? 2 : 0)
	    | (X_event.xbutton.button == Button3 ? 4 : 0);
	event.data2 = event.data3 = 0;
	H2_PostEvent(&event);
	/*  fprintf(stderr, "b"); */
	break;
      case ButtonRelease:
	event.type = ev_mouse;
	event.data1 =
	    (X_event.xbutton.state & Button1Mask)
	    | (X_event.xbutton.state & Button2Mask ? 2 : 0)
	    | (X_event.xbutton.state & Button3Mask ? 4 : 0);
	/*  suggest parentheses around arithmetic in operand of | */
	event.data1 =
	    event.data1
	    ^ (X_event.xbutton.button == Button1 ? 1 : 0)
	    ^ (X_event.xbutton.button == Button2 ? 2 : 0)
	    ^ (X_event.xbutton.button == Button3 ? 4 : 0);
	event.data2 = event.data3 = 0;
	H2_PostEvent(&event);
	/*  fprintf(stderr, "bu"); */
	break;
      case MotionNotify:
	event.type = ev_mouse;
	event.data1 =
	    (X_event.xmotion.state & Button1Mask)
	    | (X_event.xmotion.state & Button2Mask ? 2 : 0)
	    | (X_event.xmotion.state & Button3Mask ? 4 : 0);
	event.data2 = (X_event.xmotion.x - lastmousex) << 2;
	event.data3 = (lastmousey - X_event.xmotion.y) << 2;

	if (event.data2 || event.data3)
	{
	    lastmousex = X_event.xmotion.x;
	    lastmousey = X_event.xmotion.y;
	    if (X_event.xmotion.x != X_width/2 &&
		X_event.xmotion.y != X_height/2)
	    {
		H2_PostEvent(&event);
		/*  fprintf(stderr, "m"); */
		mousemoved = false;
	    } else
	    {
		mousemoved = true;
	    }
	}
	break;
	
      case Expose:
      case ConfigureNotify:
	break;
	
      default:
	if (doShm && X_event.type == X_shmeventtype) shmFinished = true;
	break;
    }

}

Cursor
createnullcursor
( Display*	display,
  Window	root )
{
    Pixmap cursormask;
    XGCValues xgc;
    GC gc;
    XColor dummycolour;
    Cursor cursor;

    cursormask = XCreatePixmap(display, root, 1, 1, 1/*depth*/);
    xgc.function = GXclear;
    gc =  XCreateGC(display, cursormask, GCFunction, &xgc);
    XFillRectangle(display, cursormask, gc, 0, 0, 1, 1);
    dummycolour.pixel = 0;
    dummycolour.red = 0;
    dummycolour.flags = 04;
    cursor = XCreatePixmapCursor(display, cursormask, cursormask,
				 &dummycolour,&dummycolour, 0,0);
    XFreePixmap(display,cursormask);
    XFreeGC(display,gc);
    return cursor;
}

/*  */
/*  I_StartTic */
/*  */
void I_StartTic (void)
{
    if (!X_display)
	return;

    while (XPending(X_display))
	I_GetEvent();

    /*  Warp the pointer back to the middle of the window */
    /*   or it will wander off - that is, the game will */
    /*   loose input focus within X11. */
    if (grabMouse)
    {
	if (!--doPointerWarp)
	{
	    XWarpPointer( X_display,
			  None,
			  X_mainWindow,
			  0, 0,
			  0, 0,
			  X_width/2, X_height/2);

	    doPointerWarp = POINTER_WARP_COUNTDOWN;
	}
    }

    mousemoved = false;
}

/*  */
/*  I_FinishUpdate */
/*  */
extern patch_t *PatchINumbers[10];

void I_Update (void)
{
   static int	lasttic;
    int		tics;
    int		i;
	int		c,d,u;

    // draws little dots on the bottom of the screen
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;

	if (tics>999) tics=999;
	c=tics / 100;
	d=(tics-c*100) / 10;
	u=(tics-c*100-d*10);

	if (PatchINumbers[c])
	{
		V_DrawPatch(320-2-24,0,PatchINumbers[c]);
		V_DrawPatch(320-2-16,0,PatchINumbers[d]);
		V_DrawPatch(320-2-8,0,PatchINumbers[u]);
	}
    }

	/* Draw image on screen */
    if (doShm)
    {

	if (!XShmPutImage(	X_display,
				X_mainWindow,
				X_gc,
				image,
				0, 0,
				0, 0,
				X_width, X_height,
				True ))
	    I_Error("XShmPutImage() failed\n");

	/*  wait for it to finish and processes all input events */
	shmFinished = false;
	do
	{
	    I_GetEvent();
	} while (!shmFinished);

    }
    else
    {

	/*  draw the image */
	XPutImage(	X_display,
			X_mainWindow,
			X_gc,
			image,
			0, 0,
			0, 0,
			X_width, X_height );

	/*  sync up with server */
	XSync(X_display, False);

    }
}


/*  */
/*  I_ReadScreen */
/*  */
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screen, SCREENWIDTH*SCREENHEIGHT*pixel_size);
}


/*  */
/*  Palette stuff. */
/*  */
static XColor	colors[256];

/*  */
/*  I_SetPalette */
/*  */
void I_SetPalette(byte *palette)
{
	int		i;
	int		c;
	byte		*palette1=palette;
	unsigned long	tc=0;
	static boolean	firstcall = true;

	if (firstcall)
	{
		firstcall = false;
		for (i=0 ; i<256 ; i++)
		{
			colors[i].pixel = i;
			colors[i].flags = DoRed|DoGreen|DoBlue;
		}
	}

	switch(pixel_size)
	{
		case 1:
			for(i = 0; i < 256; i++)
			{
				c = gammatable[usegamma][*palette++];	
				colors[i].red = (c<<8) + c;
				c = gammatable[usegamma][*palette++];
				colors[i].green = (c<<8) + c;
				c = gammatable[usegamma][*palette++];
				colors[i].blue = (c<<8) + c;
			}
			XStoreColors(X_display,X_cmap,colors,256);
			break;
		case 2:
		case 3:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	
					switch (bpp)
					{
						case 15:
							tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
							tc |= tc << 16;
							break;
						case 16:
							tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
							tc |= tc << 16;
							break;
						case 24:
							tc = (r<<16)|(v<<8)|b;
							break;
					}
					colormaps[y*256+i]=tc;			
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				switch (bpp)
				{
					case 15:
						tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
						tc |= tc << 16;
						break;
					case 16:
						tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
						tc |= tc << 16;
						break;
					case 24:
						tc = (r<<16)|(v<<8)|b;
						break;
				}
				truecolor_palette[i]=tc;
			}
			break;
	}
}

/*  */
/*  This function is probably redundant, */
/*   if XShmDetach works properly. */
/*  ddt never detached the XShm memory, */
/*   thus there might have been stale */
/*   handles accumulating. */
/*  */
void grabsharedmemory(int size)
{

  int			key = ('h'<<24) | ('x'<<16) | ('e'<<8) | 'n';
  struct shmid_ds	shminfo;
  int			minsize = 320*200;
  int			id;
  int			rc;
  /*  UNUSED int done=0; */
  int			pollution=5;
  
  /*  try to use what was here before */
  do
  {
    id = shmget((key_t) key, minsize, 0777); /*  just get the id */
    if (id != -1)
    {
      rc=shmctl(id, IPC_STAT, &shminfo); /*  get stats on it */
      if (!rc) 
      {
	if (shminfo.shm_nattch)
	{
	  fprintf(stderr, "User %d appears to be running "
		  "Hexen.  Is that wise?\n", shminfo.shm_cpid);
	  key++;
	}
	else
	{
	  if (getuid() == shminfo.shm_perm.cuid)
	  {
	    rc = shmctl(id, IPC_RMID, 0);
	    if (!rc)
	      fprintf(stderr,
		      "Was able to kill my old shared memory\n");
	    else
	      I_Error("Was NOT able to kill my old shared memory");
	    
	    id = shmget((key_t)key, size, IPC_CREAT|0777);
	    if (id==-1)
	      I_Error("Could not get shared memory");
	    
	    rc=shmctl(id, IPC_STAT, &shminfo);
	    
	    break;
	    
	  }
	  if (size >= shminfo.shm_segsz)
	  {
	    fprintf(stderr,
		    "will use %d's stale shared memory\n",
		    shminfo.shm_cpid);
	    break;
	  }
	  else
	  {
	    fprintf(stderr,
		    "warning: can't use stale "
		    "shared memory belonging to id %d, "
		    "key=0x%x\n",
		    shminfo.shm_cpid, key);
	    key++;
	  }
	}
      }
      else
      {
	I_Error("could not get stats on key=%d", key);
      }
    }
    else
    {
      id = shmget((key_t)key, size, IPC_CREAT|0777);
      if (id==-1)
      {
	extern int errno;
	fprintf(stderr, "errno=%d\n", errno);
	I_Error("Could not get any shared memory");
      }
      break;
    }
  } while (--pollution);
  
  if (!pollution)
  {
    I_Error("Sorry, system too polluted with stale "
	    "shared memory segments.\n");
    }	
  
  X_shminfo.shmid = id;
  
  /*  attach to the shared memory segment */
  image->data = X_shminfo.shmaddr = shmat(id, 0, 0);
  
  fprintf(stderr, "shared memory id=%d, addr=0x%x\n", id,
	  (int) (image->data));
}

void R_InitLightTables (void);
void R_InitColormaps (void);

void I_InitGraphics(void)
{

    char*		displayname;
    char*		d;
    int			n;
    int			pnum;
    int			x=0;
    int			y=0;

    /*  warning: char format, different type arg */
    char		xsign=' ';
    char		ysign=' ';
    
    int			oktodraw;
    unsigned long	attribmask;
    XSetWindowAttributes attribs;
    XGCValues		xgcvalues;
    int			valuemask;
    static int		firsttime=1;

    if (!firsttime)
	return;
    firsttime = 0;

    signal(SIGINT, (void (*)(int)) I_Quit);
    X_width = SCREENWIDTH ;
    X_height = SCREENHEIGHT ;

    /*  check for command-line display name */
    if ( (pnum=M_CheckParm("-disp")) ) /*  suggest parentheses around assignment */
	displayname = myargv[pnum+1];
    else
	displayname = 0;

    /*  check if the user wants to grab the mouse (quite unnice) */
    grabMouse = !!M_CheckParm("-grabmouse");

    /*  check for command-line geometry */
    if ( (pnum=M_CheckParm("-geom")) ) /*  suggest parentheses around assignment */
    {
	/*  warning: char format, different type arg 3,5 */
	n = sscanf(myargv[pnum+1], "%c%d%c%d", &xsign, &x, &ysign, &y);
	
	if (n==2)
	    x = y = 0;
	else if (n==6)
	{
	    if (xsign == '-')
		x = -x;
	    if (ysign == '-')
		y = -y;
	}
	else
	    I_Error("bad -geom parameter");
    }

    /*  open the display */
    X_display = XOpenDisplay(displayname);
    if (!X_display)
    {
	if (displayname)
	    I_Error("Could not open display [%s]", displayname);
	else
	    I_Error("Could not open display (DISPLAY=[%s])", getenv("DISPLAY"));
    }

    /*  use the default visual  */
    X_screen = DefaultScreenOfDisplay(X_display);

   	/*  Read screen depth */
	bpp = DefaultDepthOfScreen(X_screen);
	switch (bpp)
	{
		case 8:
			R_DrawColumn=R_DrawColumn8;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn8;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn8;
			R_DrawSpan=R_DrawSpan8;
			R_DrawSkyColumn=R_DrawSkyColumn8;

			R_DrawColumnLow=R_DrawColumnLow8;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow8;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow8;
			R_DrawSpanLow=R_DrawSpanLow8;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow8;

			AM_drawFline=AM_drawFline8;
			V_DrawPatch=V_DrawPatch8;
			V_DrawShadowedPatch=V_DrawShadowedPatch8;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch8;
			pixel_size=1;
			break;
		case 15:
			R_DrawColumn=R_DrawColumn16;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSkyColumn=R_DrawSkyColumn16;

			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpanLow=R_DrawSpanLow16;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow16;

			AM_drawFline=AM_drawFline16;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch16;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch16;
			pixel_size=2;
			fuzzmask=0x3DEF3DEF;
			break;
		case 16:
			R_DrawColumn=R_DrawColumn16;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSkyColumn=R_DrawSkyColumn16;

			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpanLow=R_DrawSpanLow16;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow16;

			AM_drawFline=AM_drawFline16;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch16;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch16;
			pixel_size=2;
			fuzzmask=0x7BEF7BEF;
			break;
		case 24:
			R_DrawColumn=R_DrawColumn24;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn24;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn24;
			R_DrawSpan=R_DrawSpan24;
			R_DrawSkyColumn=R_DrawSkyColumn24;

			R_DrawColumnLow=R_DrawColumnLow24;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow24;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow24;
			R_DrawSpanLow=R_DrawSpanLow24;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow24;

			AM_drawFline=AM_drawFline24;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch24;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch24;
			pixel_size=3;
			break;
		default:
			I_Error("Unknown pixel size for screen\n");
	}


  if(bpp==8) {
    if(!XMatchVisualInfo(X_display, DefaultScreen(X_display),
			 bpp, PseudoColor, &X_visualinfo)) {
      I_Error("Could not get PseudoColor visual!\n");
    }
  } else {
    if(!XMatchVisualInfo(X_display, DefaultScreen(X_display),
			 bpp, TrueColor, &X_visualinfo))
      I_Error("Could not get TrueColor visual!\n");
  }
    X_visual = X_visualinfo.visual;

    /*  check for the MITSHM extension */
    doShm = XShmQueryExtension(X_display);

    /*  even if it's available, make sure it's a local connection */
    if (doShm)
    {
	if (!displayname) displayname = (char *) getenv("DISPLAY");
	if (displayname)
	{
	    d = displayname;
	    while (*d && (*d != ':')) d++;
	    if (*d) *d = 0;
	    if (!(!strcasecmp(displayname, "unix") || !*displayname)) doShm = false;
	}
    }

    fprintf(stderr, "Using MITSHM extension\n");

    /*  create the colormap */
	if (bpp==8)
	{
		X_cmap = XCreateColormap(	X_display,
						RootWindowOfScreen(X_screen),
						X_visual,
						AllocAll);
	} else {
	  	X_cmap = DefaultColormapOfScreen(X_screen);
	}

    /*  setup attributes for main window */
    attribmask = CWEventMask | CWColormap | CWBorderPixel;
    attribs.event_mask =
	KeyPressMask
	| KeyReleaseMask
	/*  | PointerMotionMask | ButtonPressMask | ButtonReleaseMask */
	| ExposureMask;

    attribs.colormap = X_cmap;
    attribs.border_pixel = 0;

    /*  create the main window */
    X_mainWindow = XCreateWindow(	X_display,
					RootWindowOfScreen(X_screen),
					x, y,
					X_width, X_height,
					0, /*  borderwidth */
					bpp, /*  depth */
					InputOutput,
					X_visual,
					attribmask,
					&attribs );

    XDefineCursor(X_display, X_mainWindow,
		  createnullcursor( X_display, X_mainWindow ) );

    /*  create the GC */
    valuemask = GCGraphicsExposures;
    xgcvalues.graphics_exposures = False;
    X_gc = XCreateGC(	X_display,
  			X_mainWindow,
  			valuemask,
  			&xgcvalues );

    /*  map the window */
    XMapWindow(X_display, X_mainWindow);

    /*  wait until it is OK to draw */
    oktodraw = 0;
    while (!oktodraw)
    {
	XNextEvent(X_display, &X_event);
	if (X_event.type == Expose
	    && !X_event.xexpose.count)
	{
	    oktodraw = 1;
	}
    }

    /*  grabs the pointer so it is restricted to this window */
    if (grabMouse)
	XGrabPointer(X_display, X_mainWindow, True,
		     ButtonPressMask|ButtonReleaseMask|PointerMotionMask,
		     GrabModeAsync, GrabModeAsync,
		     X_mainWindow, None, CurrentTime);

    if (doShm)
    {

	X_shmeventtype = XShmGetEventBase(X_display) + ShmCompletion;

	/*  create the image */
	image = XShmCreateImage(	X_display,
					X_visual,
					bpp,
					ZPixmap,
					0,
					&X_shminfo,
					X_width,
					X_height );

	grabsharedmemory(image->bytes_per_line * image->height);


	/*  UNUSED */
	/*  create the shared memory segment */
	/*  X_shminfo.shmid = shmget (IPC_PRIVATE, */
	/*  image->bytes_per_line * image->height, IPC_CREAT | 0777); */
	/*  if (X_shminfo.shmid < 0) */
	/*  { */
	/*  perror(""); */
	/*  I_Error("shmget() failed in InitGraphics()"); */
	/*  } */
	/*  fprintf(stderr, "shared memory id=%d\n", X_shminfo.shmid); */
	/*  attach to the shared memory segment */
	/*  image->data = X_shminfo.shmaddr = shmat(X_shminfo.shmid, 0, 0); */
	

	if (!image->data)
	{
	    perror("");
	    I_Error("shmat() failed in InitGraphics()");
	}

	/*  get the X server to attach to it */
	if (!XShmAttach(X_display, &X_shminfo))
	    I_Error("XShmAttach() failed in InitGraphics()");

    }
    else
    {
	image = XCreateImage(	X_display,
    				X_visual,
    				bpp,
    				ZPixmap,
    				0,
    				(char*)malloc(X_width * X_height * pixel_size),
    				X_width, X_height,
    				bpp,
    				X_width * pixel_size);

    }

	screen = (unsigned char *) (image->data);

	/*  Init palette */
	I_SetPalette(W_CacheLumpName("PLAYPAL",PU_CACHE));
}
